// ----------------------------- SMSTCPGateway -----------------------------------------------------------------------------------------------
//
// This sketch is designed to work with the Haute Solutions EGGBoard with a Haute Solutions OLED Panel
// (www.haute-solutions.com)
//
// Haute Solutions Libraries Required: SMSDEV ( GSMSMS, OLED_SSD1311 )
//
// Eggboard should be configured to have shield pins D4, D6, D7, and D10 serviced from Arduino pins 
// D46, D47, D48, and D49 respectively (redirected)
//
// Eggboard should also be configured for 3.3V signalling!
//
// This sketch will listen for a TCP connection on the specified port and then convert and send that message
// as an SMS text message to the designated SMS Phone number.  Inbound message format on TCP port should be
// multiple lines of text formatted as follows:
//
//      {SenderIP}
//      SMS_SEND
//      {SMS Recipient Number}
//      {Message Line1}   
//      {Message Line2}      
//      {Message Line...}   
//
//      For Example:
//
//      192.168.0.1
//      SMS_SEND
//      +12345678900
//      This is a multiline
//      message!
//
// This sketch will also listen for inbound SMS messages which will then be sent to the specified port on a
// Server for additional processing/handling.  SMS Messages will be reformatted as follows:
//
//      {DeviceIP}
//      SMS_RCV
//      {SMS Sender Number}
//      {Message Line1}   
//      {Message Line2}      
//      {Message Line...}   
//
//      For Example:
//
//      192.168.0.2
//      SMS_RCV
//      +12345678900
//      This is a multiline
//      message!
//
// IP Address and Port of Server to handle/process inbound SMS messages must be specified.
//
// OLED Connections:
//                          GND to GND
//                          VCC to 3V3
//                          SDA to SDA (20)
//                          SCL to SCL (21)
//
// Button Pad Connections:  (Ensure Pins 4, 6, and 7 are redirected from D46, D47, and D48 on EGGBoard PCB!!!)
//                          B1 to 4
//                          B2 to 5
//                          B3 to 6
//                          B4 to 7
//                          GND to GND
//
// This file is free software; you can redistribute it and/or modify
// it under the terms of either the GNU General Public License version 2
// or the GNU Lesser General Public License version 2.1, both as
// published by the Free Software Foundation.
//
// --------------------------------------------------------------------------------------------------------------------------
#include "SMSDEV.h"
#include <Ethernet.h>

#define COMPANY   "Haute Solutions"
#define DEVICE    "EGGBox"
#define VERSION   "V6.3"
#define EEPROMSIG "SMSTCPGateway 2.0"     // EEPROM Valid Signature (19 Char Max)

// NOTE:  Ensure Pins 4, 6, and 7 are redirected from D46, D47, and D48 on EGGBoard PCB!!!
#define BOARD EGGBOARD3G                  // Compile for EGGBoard3G (Mega2560)
#define OLEDADDR 0x3C                     // I2C Address for OLED Display (NULL if None)
#define BUTTON1_PIN 46                    // Display Main Screen  (PCB 4)           
#define BUTTON2_PIN 5                     // Display NetInfo      (PCB 5)
#define BUTTON3_PIN 47                    // Display CfgInfo      (PCB 6)
#define BUTTON4_PIN 48                    // Factory Reset        (PCB 7)

//#define BUTTON1_PIN 5                     // Display Main Screen             
//#define BUTTON2_PIN 11                    // Display NetInfo
//#define BUTTON3_PIN 12                    // Display CfgInfo
//#define BUTTON4_PIN 13                    // Factory Reset
// SDA = 20
// SCL = 21

CMDINFO Commands[] = {                    // Device Specific Commands (Generic SMS Commands are defined in SMSDEV Library)
  {"HELP",         1, 0},
  {"STATUS",       2, 0},
  {"MACADDR:",     3, 1},
  {"EGGIP:",       4, 1},
  {"SERVERIP:",    5, 1},
  {"NOTIFY:",      6, 1}
};

//--------------------------------------------------------------------------------------------------------------------------------------
//  The following global parameters are set (initially) in the Site_Data.h file and 
//  maintained in EEPROM so they are persistant and can be remotely updated.
char EGGMAC[EEFIELDSIZE];            // MAC Address for the EGGBridge
char EGGIP[EEFIELDSIZE];             // IP Address for the EGGBridge (0.0.0.0 for DHCP)
char ServerIP[EEFIELDSIZE];          // IP Address of Server for processing SMS Messages                                                     
char Notify[EEFIELDSIZE];            // SMS Recipient for Heartbeat Notification
char Password[EEFIELDSIZE];          // Password to authorize a Device
//--------------------------------------------------------------------------------------------------------------------------------------

// Define some global shorthand "aliases" (for readability)
char* Buffer;
char* Caller;
char* Temp;
char* Param;
char* DevName;
char* DevNum;

SMSDEV SMSDev(BOARD, OLEDADDR);

#include "Site_Data.h"

// Inbound TCP port to listen on
// Currently, Ethernet Library does not allow us to set this later (i.e. Setup())
EthernetServer TCPServer(ServicePort);
EthernetClient client;

//  The following parameters are loaded from EEPROM or Site_Data.h as appropriate
byte EGGIPAddr[] = { 0, 0, 0, 0 };                            // EGGBoard IP Address
byte EGGMACAddr[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };   // EGGBoard MAC Address
IPAddress ServerIPAddr( 0, 0, 0, 0 );                         // IP Address of the Server to process SMS Messages

#define TCPBUFFSIZE 512
char TCPBuffer[TCPBUFFSIZE];

void setup(){
int rcode;

  Serial.begin(9600);

  // Uncomment the line below to Wack the EEPROM Signature and force a factory reset
  // EEPROM.write(0, '\0');
  
  SetButtons();

  rcode = SMSDev.Setup(COMPANY, DEVICE, VERSION, EEPROMSIG, "#", Commands, sizeof(Commands)/sizeof(Commands[0]));
  while (rcode > 0) { }
  
  SMSDev.SetPrivProcessing(0);      // Disable Priv Command Checking

  // Assign our Global Aliases
  Buffer       = SMSDev.Buffer;
  Caller       = SMSDev.Caller;
  Temp         = SMSDev.Temp;
  Param        = SMSDev.Param;
  DevName      = SMSDev.DevName;
  DevNum       = SMSDev.DevNum;

  if(!GetDefaults()) SetDefaults();
  
  MACFromString(EGGMAC, EGGMACAddr);
  IPFromString(EGGIP, EGGIPAddr);
  IPFromString(ServerIP, &ServerIPAddr);
  StartNet();

  if(SMSDev.GetAuthCaller(0, Caller) != 0) {
    if(SMSDev.BootNotify[0] == '1') {
      CommandSTATUS(Caller, "System Ready!");
    }
  }

  SendTCPMessage("HELLO", "EGGBRIDGE", NULL);

  //Process any pending/stored SMS Messages
  while (SMSDev.CheckSMS() == 1) {
  }

  SMSDev.DisplayIdleScreen();
}

void loop(){
    
  switch (SMSDev.Loop()) {
    case -7:  // CFGINFO (SMSDEV) = Add additional STATUS Screen
      CommandSTATUS(Caller, NULL);
      break;
    case -9:  // Factory Reset (SMSDEV)
      CommandRESET();
      break;

    case 1:   // HELP
      CommandHELP();
      break;
    case 2:   // STATUS
      CommandSTATUS(Caller, NULL);
      break;
    case 3:   // MACADDR:
      SetConfig(Param, EGGMAC, EEDEVSPEC1, "MAC Addr Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 4:   // EGGIP:
      SetConfig(Param, EGGIP, EEDEVSPEC2, "IP Addr Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 5:   // SERVERIP:
      SetConfig(Param, ServerIP, EEDEVSPEC3, "Server IP Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 6:   // NOTIFY:
      SetConfig(Param, Notify, EEAUTHCALLERS, "Notification Number Set!", Caller);
      break;

    case 100: // Clock Tick (every 30 sec)
      break;
    case 101: // Alarm 1
      break;
    case 102: // Alarm 2
      break;
    case 103: // HeartBeat
      if(SMSDev.GetAuthCaller(0, Caller) != 0) CommandSTATUS(Caller, "HeartBeat!");
      break;
    case 104: // CSQ (Signal Change)
      // Display Updated in SMSDEV
      break;
    case 200: // Unauthorized Caller 
      // Message Forwarded in SMSDEV
      break;
    case 201: // Invalid Command
      // Reply Handled in SMSDEV
      break;
    case 202: // Command Restricted
      // Reply Handled in SMSDEV
      break;
    case 203: // Forwarded Message for local Processing (Not a Command)
      SendTCPMessage("SMS_RCV", Caller, Buffer);
      break;
  }

  CheckTCP();                                 // Any TCP Messages Pending?

  CheckButtons();
}

// ------------------------------ TCP Support Functions -----------------------------------------

void StartNet() {
int i;
int DHCPStat;
char Sequence[17];

  SMSDev.DisplayWithDelay("Starting", "Ethernet", 1);
  
  for(i=0; i<6; i++) {
    Temp[i*2] = EGGMAC[i*3];
    Temp[i*2+1] = EGGMAC[i*3+1];
  }
  Temp[12]='\0';
  SMSDev.DisplayWithDelay("MAC Address", Temp, 1);

  if(strcmp(EGGIP, "0.0.0.0") != 0) {
    Ethernet.begin(EGGMACAddr, EGGIPAddr);
  } else {
    SMSDev.DisplayWithDelay("Performing DHCP", "");
    if(Ethernet.begin(EGGMACAddr) == 0) {
      i=0;
      do {
        if(i > 16) i=0;
        strcpy(Sequence, "                ");
        Sequence[i] = '>';
        i++;
        SMSDev.Display(NULL, Sequence); 
        delay(1000);
      
        switch (Ethernet.maintain()) {
          case 0:   // No response
            DHCPStat=0;
            break;
          case 1:   // Renew Failed
            DHCPStat=0;
            break;
          case 2:   // Renew Sucess
            DHCPStat=1;
            break;
          case 3:   // Rebind Failed
            DHCPStat=0;
            break;
          case 4:   // Rebind Success
            DHCPStat=1;
            break;
        }
      } while (DHCPStat == 0);
    }
    IPAddress LocalIP = Ethernet.localIP();
    sprintf(EGGIP, "%d.%d.%d.%d", LocalIP[0],LocalIP[1],LocalIP[2],LocalIP[3]);
  }

  SMSDev.DisplayWithDelay("IP Address", EGGIP, 1);

  SMSDev.DisplayWithDelay("Server Address", ServerIP, 1);

  SMSDev.DisplayWithDelay("Service Port", itoa(ServicePort, Temp, 10), 1);

}

void CheckTCP() {  
char c;
int offset;
unsigned long timeout;

EthernetClient client;

  client = TCPServer.available();
  if (!client) return;

  TCPBuffer[0] = '\0';
  offset=0;
    
  while (client.connected()) {
    if (client.available()) {
        
      c = client.read();
      timeout = millis() + 1000;
      if(offset < (TCPBUFFSIZE-2)) {
        TCPBuffer[offset++] = c;
        TCPBuffer[offset] ='\0';
      }
      
    } else {
      if(millis() > timeout) {
        client.println("OK");
        break;
      }
    }   
  }

  client.stop();
  ProcessTCPMessage(TCPBuffer);
  Serial.println();
}

int ProcessTCPMessage(char *Cmd) {
char *p1;
char *p2;
char *p3;
char *p4;
char *p5;

  Serial.println("TCP Incoming:");
  Serial.println(Cmd);
  Serial.println("");

  p1=Cmd;
  p2=NULL;
  p3=NULL;
  p4=NULL;

  p2=strchr(p1,'\n');
  if(p2 != NULL) {
    *p2='\0';
    p2++;

    p3=strchr(p2,'\n');
    if(p3 != NULL) {
      *p3='\0';
      p3++;

      p4=strchr(p3,'\n');
      if(p4 != NULL) {
        *p4='\0';
        p4++;

      }
    }
  }

  if(!strcasecmp(p2,"SMS_SEND")) {
    Serial.print("Sending SMS to ");
    Serial.print(p3);
    Serial.println("");
    Serial.println(p4);
    Serial.println("");
    SMSDev.SendSMSWithDisplay(p3, p4);
    
  } else if(!strcasecmp(p2,"DISPLAY")) {
    p5=strchr(p4,'\n');
    if(p5 != NULL) *p5='\0';
    Serial.print("Display:");
    Serial.print(p3);
    Serial.println(p4);
    SMSDev.Display(p3,p4);

  } else if(!strcasecmp(p2,"CLEAR")) {
    Serial.print("Clearing Display");
    SMSDev.Display("","");
    
  } else {
    Serial.println("(Ignored)");
    
  }
  Serial.println();
}

int SendTCPMessage(char *p1, char *p2, char *p3) {
  
  strcpy(TCPBuffer,EGGIP);
  strcat(TCPBuffer,"\n");
  strcat(TCPBuffer,p1);
  if(p2 != NULL) {
    strcat(TCPBuffer,"\n");
    strcat(TCPBuffer,p2);
  }
  if(p3 != NULL) {
    strcat(TCPBuffer,"\n");
    strcat(TCPBuffer,p3);;
  }

  Serial.println();
  Serial.print("Message to: ");
  Serial.print(ServerIPAddr);
  Serial.print(" (Port: ");
  Serial.print(ServicePort);
  Serial.println(")");

  Serial.println(TCPBuffer);
 
  if (client.connect(ServerIPAddr,ServicePort)) {
    
    client.write(TCPBuffer);
//  client.write("\r\n");

    Serial.println("(Forwarded)");
    Serial.println();
    client.stop();
    return 1;
    
  } else {
    
    Serial.println("(Failed)");
    Serial.println();
    client.stop();
    return 0;
  }
}

// ---------------------------- Display Functions ---------------------------------------

void DisplayAllInfo() {

    SMSDev.DisplayTime();
    DisplayNetInfo();
    SMSDev.DisplayCfgInfo();
    SMSDev.AuthCallers = SMSDev.DisplayAuthCallers();
}

void DisplayNetInfo() {

  for(int i=0; i<6; i++) {
    Temp[i*2] = EGGMAC[i*3];
    Temp[i*2+1] = EGGMAC[i*3+1];
  }
  Temp[12]='\0';
  SMSDev.DisplayWithDelay("MAC Address", Temp, 1);

  SMSDev.DisplayWithDelay("IP Address", EGGIP, 1);
  
  SMSDev.DisplayWithDelay("Server Address", ServerIP, 1);

  SMSDev.DisplayWithDelay("Service Port", itoa(ServicePort, Temp, 10), 1);
}

// ---------------------------- Button Functions ---------------------------------------

void SetButtons() {

  pinMode(BUTTON1_PIN, INPUT_PULLUP);
  pinMode(BUTTON2_PIN, INPUT_PULLUP);
  pinMode(BUTTON3_PIN, INPUT_PULLUP);
  pinMode(BUTTON4_PIN, INPUT_PULLUP);  
}

void CheckButtons() {

  if(digitalRead(BUTTON1_PIN) == LOW) {
    delay(500); // Debounce
    SMSDev.DisplayIdleScreen();
  }
  
  if(digitalRead(BUTTON2_PIN) == LOW) {
    delay(500); // Debounce
    DisplayAllInfo();
    SMSDev.DisplayIdleScreen();
  }
  
  if(digitalRead(BUTTON3_PIN) == LOW) {
    delay(500); // Debounce
    if (DisplayCountdown(BUTTON3_PIN, "Requesting HELP")) {
      strcpy(Buffer, "ASSISTANCE REQUESTED!");
      if(strlen(SMSDev.DevName)) {
        strcat(Buffer, "\nfrom ");
        strcat(Buffer, SMSDev.DevName);
      }
      SMSDev.SendSMSWithDisplay(Notify, Buffer); 
    }
  }
  
  if(digitalRead(BUTTON4_PIN) == LOW) {
    delay(500); // Debounce
    if(DisplayCountdown(BUTTON4_PIN, "Factory Reset")) {
      SMSDev.FactoryReset();
      CommandRESET();
    }
  }
}

int DisplayCountdown(int Button, char* Msg) {
unsigned long ResetTimer;
char ALine[20];
int SCheck;
int SecsRemaining = -1;
int SecsToWait = 5;

  ResetTimer = millis();
  do {
    SCheck = SecsToWait - ((millis() - ResetTimer) / 1000);
    if(SecsRemaining != SCheck) {
      SecsRemaining = SCheck;
      sprintf(ALine,"in %d Seconds...", SecsRemaining);
      SMSDev.Display(Msg, ALine);
    }
    if(SecsRemaining == 0) return 1;
    
  } while (digitalRead(Button) == LOW);

  SMSDev.DisplayIdleScreen();

  return 0;
}


// ---------------------------- Command Processing Functions ---------------------------------------

void CommandRESET() {
  // Majority of Factory Reset handled by SMSDEV (inc Notification)
  // Do any Local Resets here...
  SetDefaults();
}

void CommandHELP() {
  
  strcpy(Buffer, DEVICE);
  strcat(Buffer, " Commands:\n\n");
  
  strcat(Buffer, "STATUS\n\n");
  strcat(Buffer, "(Other Config CMDs require Auth)\n\n");

  SMSDev.SendSMSWithDisplay(Caller,Buffer);  
}

void CommandSTATUS(char* acaller, char *msg) {
char csqval[2];

  strcpy(Buffer,"");

  if(msg != NULL) {
    strcpy(Buffer, msg);
    strcat(Buffer, "\n\n");
  }

  strcat(Buffer, DEVICE);
  strcat(Buffer, ": ");
  strcat(Buffer, VERSION);
  strcat(Buffer, "\n");
    
  strcat(Buffer, "Time: ");
  strcat(Buffer, SMSDev.TimeNow);
  strcat(Buffer, "\n");

  csqval[0]=char(SMSDev.CSQ + 48);
  csqval[1]='\0';
  strcat(Buffer, "Signal: ");
  strcat(Buffer, csqval);
  strcat(Buffer, "/5");
  strcat(Buffer, "\n");

  sprintf(Temp, "%02X:%02X:%02X:%02X:%02X:%02X", EGGMACAddr[0],EGGMACAddr[1],EGGMACAddr[2],EGGMACAddr[3],EGGMACAddr[4],EGGMACAddr[5]);
  strcat(Buffer, "MAC: ");
  strcat(Buffer, Temp);
  strcat(Buffer, "\n");

  strcat(Buffer, "IPAddress: ");
  strcat(Buffer, EGGIP);
  strcat(Buffer, "\n");

  strcat(Buffer, "Service Port: ");
  strcat(Buffer, itoa(ServicePort,Temp,10));
  strcat(Buffer, "\n");

  strcat(Buffer, "Server IP: ");
  strcat(Buffer, ServerIP);
  strcat(Buffer, "\n");

  SMSDev.SendSMSWithDisplay(Caller,Buffer);  
}

// ------------------------- Setup and Configuration --------------------------------------------

void SetConfig(char *Value,  char *Variable, int EEFieldNo, char *ReplyMsg, char *ReplyTo) {
  strncpy(Variable, Value, EEFIELDSIZE);
  Variable[EEFIELDSIZE-1] = '\0';
  SMSDev.WriteEEPROMString(EEFieldNo, Variable, EEFIELDSIZE);
  SMSDev.SendSMSWithDisplay(ReplyTo, ReplyMsg);  
}

int GetDefaults() {

  SMSDev.ReadEEPROMField(EEDEVSPEC1, EGGMAC);
  if(strlen(EGGMAC) == 0) return 0;

  SMSDev.ReadEEPROMField(EEDEVSPEC2, EGGIP);
  SMSDev.ReadEEPROMField(EEDEVSPEC3, ServerIP);
  SMSDev.ReadEEPROMField(EEPWD,      Password);
  SMSDev.ReadEEPROMField(EEDEVNAME,  DevName);
  SMSDev.ReadEEPROMField(EEDEVNUM,   DevNum);
  
  SMSDev.GetAuthCaller(0, Notify);  
  return 1;
}

void SetDefaults() {

// OPTIONAL:  May be set via SMS Command
  SiteDefaults();

  SMSDev.WriteEEPROMField(EEDEVSPEC1, EGGMAC);
  SMSDev.WriteEEPROMField(EEDEVSPEC2, EGGIP);
  SMSDev.WriteEEPROMField(EEDEVSPEC3, ServerIP);
  SMSDev.WriteEEPROMField(EEPWD,      Password);
  SMSDev.WriteEEPROMField(EEDEVNAME,  DevName);
  SMSDev.WriteEEPROMField(EEDEVNUM,   DevNum);

  SMSDev.Reboot();
}

// ----------------------------- MAC / IP Conversion Functions -----------------------------------------------------

int MACFromString(char *macstr, byte *macaddr) {
int i;
char swork[18];
char *p1;
char *p2;

  if(CheckMACString(macstr)) return 0;

  strcpy(swork,macstr);
  strupr(swork);
  
  p1=swork;
  for(i=0; i<5; i++){
    p2=strchr(p1,':');
   *p2='\0';
    macaddr[i]=strtol(p1,NULL,16);
    p1=p2+1;
  }
  macaddr[5]=strtol(p1,NULL,16);
  
  return 1;
}

int IPFromString(char *ipstr, IPAddress *ipaddr) {
int i;
char swork[16];
char *p1;
char *p2;

  if(CheckIPString(ipstr)) return 0;

  strcpy(swork,ipstr);
  p1=swork;
  for(i=0; i<3; i++){
    p2=strchr(p1,'.');
   *p2='\0';
    (*ipaddr)[i]=atoi(p1);
    p1=p2+1;
  }
  (*ipaddr)[3]=atoi(p1);
  
  return 1;
}

int IPFromString(char *ipstr, byte *ipaddr) {
int i;
char swork[16];
char *p1;
char *p2;

  if(CheckIPString(ipstr)) return 0;

  strcpy(swork,ipstr);
  p1=swork;
  for(i=0; i<3; i++){
    p2=strchr(p1,'.');
   *p2='\0';
    ipaddr[i]=atoi(p1);
    p1=p2+1;
  }
  ipaddr[3]=atoi(p1);
  
  return 1;
}

int CheckMACString(char *macstr) {
int i;
int c;
int slen;
char swork[18];

  slen=strlen(macstr);
  if(slen != 17) return 3;

  strcpy(swork,macstr);
  strupr(swork);
  
  for(i=0; i<slen; i++){
    c=(int)swork[i];
    if(c < 48) return 1;
    if((c > 58) && (c < 65)) return 1;
    if(c > 70) return 1;
  }

  for(i=2; i<slen; i=i+3){
    if(swork[i] != ':') return 2;
  }

  return 0;
}

int CheckIPString(char *ipstr) {
int i;
int slen;
int pcount;
int dcount;

  pcount=0;
  dcount=0;
  slen=strlen(ipstr);
  for(i=0; i<slen; i++){
    if(isDigit(ipstr[i])) dcount++;
    if(ipstr[i] == '.') pcount++;
  }
  if(pcount != 3) return 1;
  if((pcount+dcount) != slen) return 2;
  if((pcount+dcount) > 15) return 3;
  if((pcount+dcount) < 7) return 4;

  return 0;
}
